"""add_creator_in_vfolders

Revision ID: 4b7b650bc30e
Revises: 202b6dcbc159
Create Date: 2019-10-30 00:23:57.085692

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4b7b650bc30e"
down_revision = "202b6dcbc159"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("vfolders", sa.Column("creator", sa.String(length=128), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("vfolders", "creator")
    # ### end Alembic commands ###



"""stringify_userid

Revision ID: f8a71c3bffa2
Revises: bf4bae8f942e
Create Date: 2018-06-17 13:52:13.346856

"""
import os

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import convention

# revision identifiers, used by Alembic.
revision = "f8a71c3bffa2"
down_revision = "bf4bae8f942e"
branch_labels = None
depends_on = None


def upgrade():
    metadata = sa.MetaData(naming_convention=convention)
    keypairs = sa.Table(
        "keypairs",
        metadata,
        sa.Column("user_id", sa.String(length=256), index=True),
    )

    print("Choose keypairs.user_id column migrate option:")
    print(" [a] Convert all numeric user IDs to strings directly")
    print(
        " [b] Convert numeric user IDs to strings using a mapping table\n"
        "     (user_id_map.txt must be present in the current working directory\n"
        "      which contains a space-sep.list of numeric and string ID pairs.)"
    )
    print("NOTE: If you choose [b], you will not be able to downgrade!")

    choice = os.environ.get("STRINGIFY_USERID_CHOICE")
    if choice is None:
        while True:
            choice = input("Your choice? [a/b] ")
            if choice in ("a", "b"):
                break
            print("Invalid choice.")
            continue

    op.alter_column("keypairs", "user_id", existing_type=sa.Integer(), type_=sa.String(length=256))

    # NOTE: We do the data migration after converting column type.

    if choice == "b":
        # query all unique user ids
        q = sa.select([keypairs.c.user_id]).group_by(keypairs.c.user_id)
        rows = op.get_bind().execute(q)
        user_ids = set(int(row.user_id) for row in rows)
        print(f"There are {len(user_ids)} unique user IDs.")

        user_id_map = {}
        with open("user_id_map.txt", "r") as f:
            for line in f:
                num_id, str_id = line.split(maxsplit=1)
                assert len(str_id) <= 256, f"Too long target user ID! ({num_id} -> {str_id!r})"
                user_id_map[int(num_id)] = str_id

        map_diff = user_ids - set(user_id_map.keys())
        assert len(map_diff) == 0, f"There are unmapped user IDs!\n{map_diff}"

        for num_id, str_id in user_id_map.items():
            op.execute(
                keypairs.update()
                .values({"user_id": str_id})
                .where(keypairs.c.user_id == str(num_id))
            )


def downgrade():
    op.alter_column("keypairs", "user_id", existing_type=sa.Integer(), type_=sa.String(length=256))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import click
import graphene

from ai.backend.common.logging import BraceStyleAdapter

from ..models.gql import Mutations, Queries

if TYPE_CHECKING:
    from .context import CLIContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@click.group()
def cli(args) -> None:
    pass


@cli.command()
@click.pass_obj
def show(cli_ctx: CLIContext) -> None:
    with cli_ctx.logger:
        schema = graphene.Schema(query=Queries, mutation=Mutations, auto_camelcase=False)
        log.info("======== GraphQL API Schema ========")
        print(str(schema))



"""update_cluster_columns_in_kernels

Revision ID: 548cc8aa49c8
Revises: 1e673659b283
Create Date: 2020-09-08 18:50:05.594899

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "548cc8aa49c8"
down_revision = "1e673659b283"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("ix_kernels_sess_id_role", table_name="kernels")
    op.drop_index("ix_kernels_unique_sess_token", table_name="kernels")

    op.add_column(
        "kernels",
        sa.Column(
            "cluster_size", sa.Integer, nullable=False, default=1, server_default=sa.text("1")
        ),
    )
    op.add_column("kernels", sa.Column("cluster_hostname", sa.String(length=64), nullable=True))
    conn = op.get_bind()
    query = (
        "UPDATE kernels k "
        "  SET cluster_size = (SELECT COUNT(*) FROM kernels j WHERE j.session_id = k.session_id);"
    )
    conn.execute(text(query))
    query = "UPDATE kernels SET cluster_hostname = CONCAT(role, CAST(idx AS TEXT));"
    conn.execute(text(query))
    op.alter_column("kernels", "cluster_hostname", nullable=False)

    op.alter_column("kernels", "idx", new_column_name="cluster_idx", nullable=False)
    op.alter_column("kernels", "role", new_column_name="cluster_role", nullable=False)

    op.create_index(
        "ix_kernels_sess_id_role", "kernels", ["session_id", "cluster_role"], unique=False
    )
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_id"],
        unique=True,
        postgresql_where=sa.text(
            "status NOT IN ('TERMINATED', 'CANCELLED') " "and cluster_role = 'main'"
        ),
    )


def downgrade():
    op.drop_index("ix_kernels_unique_sess_token", table_name="kernels")
    op.drop_index("ix_kernels_sess_id_role", table_name="kernels")

    op.alter_column("kernels", "cluster_idx", new_column_name="idx")
    op.alter_column("kernels", "cluster_role", new_column_name="role")
    op.drop_column("kernels", "cluster_size")
    op.drop_column("kernels", "cluster_hostname")

    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_name"],
        unique=True,
        postgresql_where=sa.text("status NOT IN ('TERMINATED', 'CANCELLED') " "and role = 'main'"),
    )
    op.create_index("ix_kernels_sess_id_role", "kernels", ["session_name", "role"], unique=False)



from __future__ import annotations

import datetime as dt
import logging
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, MutableMapping, Tuple

import aiohttp_cors
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web

from ai.backend.common import redis_helper
from ai.backend.common import validators as tx
from ai.backend.common.distributed import GlobalTimer
from ai.backend.common.events import AbstractEvent, EmptyEventArgs, EventHandler
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import AgentId, LogSeverity, RedisConnectionInfo

from ..defs import REDIS_LIVE_DB, LockID
from ..models import UserRole
from ..models import association_groups_users as agus
from ..models import error_logs, groups
from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class DoLogCleanupEvent(EmptyEventArgs, AbstractEvent):
    name = "do_log_cleanup"


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict(
        {
            t.Key("severity"): tx.Enum(LogSeverity),
            t.Key("source"): t.String,
            t.Key("message"): t.String,
            t.Key("context_lang"): t.String,
            t.Key("context_env"): tx.JSONString,
            t.Key("request_url", default=None): t.Null | t.String,
            t.Key("request_status", default=None): t.Null | t.Int,
            t.Key("traceback", default=None): t.Null | t.String,
        },
    )
)
async def append(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    params["domain"] = request["user"]["domain_name"]
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    requester_uuid = request["user"]["uuid"]
    log.info(
        "CREATE (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )

    async with root_ctx.db.begin() as conn:
        resp = {
            "success": True,
        }
        query = error_logs.insert().values(
            {
                "severity": params["severity"],
                "source": params["source"],
                "user": requester_uuid,
                "message": params["message"],
                "context_lang": params["context_lang"],
                "context_env": params["context_env"],
                "request_url": params["request_url"],
                "request_status": params["request_status"],
                "traceback": params["traceback"],
            }
        )
        result = await conn.execute(query)
        assert result.rowcount == 1
    return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict(
        {
            t.Key("mark_read", default=False): t.ToBool(),
            t.Key("page_size", default=20): t.ToInt(lt=101),
            t.Key("page_no", default=1): t.ToInt(),
        }
    ),
)
async def list_logs(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    resp: MutableMapping[str, Any] = {"logs": []}
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info(
        "LIST (ak:{0}/{1})",
        requester_access_key,
        owner_access_key if owner_access_key != requester_access_key else "*",
    )
    async with root_ctx.db.begin() as conn:
        is_admin = True
        select_query = (
            sa.select([error_logs])
            .select_from(error_logs)
            .order_by(sa.desc(error_logs.c.created_at))
            .limit(params["page_size"])
        )
        count_query = sa.select([sa.func.count()]).select_from(error_logs)
        if params["page_no"] > 1:
            select_query = select_query.offset((params["page_no"] - 1) * params["page_size"])
        if request["is_superadmin"]:
            pass
        elif user_role == UserRole.ADMIN or user_role == "admin":
            j = groups.join(agus, groups.c.id == agus.c.group_id)
            usr_query = (
                sa.select([agus.c.user_id])
                .select_from(j)
                .where(groups.c.domain_name == domain_name)
            )
            result = await conn.execute(usr_query)
            usrs = result.fetchall()
            user_ids = [g.id for g in usrs]
            where = error_logs.c.user.in_(user_ids)
            select_query = select_query.where(where)
            count_query = count_query.where(where)
        else:
            is_admin = False
            where = (error_logs.c.user == user_uuid) & (~error_logs.c.is_cleared)
            select_query = select_query.where(where)
            count_query = count_query.where(where)

        result = await conn.execute(select_query)
        for row in result:
            result_item = {
                "log_id": str(row["id"]),
                "created_at": datetime.timestamp(row["created_at"]),
                "severity": row["severity"],
                "source": row["source"],
                "user": row["user"],
                "is_read": row["is_read"],
                "message": row["message"],
                "context_lang": row["context_lang"],
                "context_env": row["context_env"],
                "request_url": row["request_url"],
                "request_status": row["request_status"],
                "traceback": row["traceback"],
            }
            if result_item["user"] is not None:
                result_item["user"] = str(result_item["user"])
            if is_admin:
                result_item["is_cleared"] = row["is_cleared"]
            resp["logs"].append(result_item)
        resp["count"] = await conn.scalar(count_query)
        if params["mark_read"]:
            read_update_query = (
                sa.update(error_logs)
                .values(is_read=True)
                .where(error_logs.c.id.in_([x["log_id"] for x in resp["logs"]]))
            )
            await conn.execute(read_update_query)
        return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
async def mark_cleared(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    domain_name = request["user"]["domain_name"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    log_id = uuid.UUID(request.match_info["log_id"])

    log.info("CLEAR")
    async with root_ctx.db.begin() as conn:
        update_query = sa.update(error_logs).values(is_cleared=True)
        if request["is_superadmin"]:
            update_query = update_query.where(error_logs.c.id == log_id)
        elif user_role == UserRole.ADMIN or user_role == "admin":
            j = groups.join(agus, groups.c.id == agus.c.group_id)
            usr_query = (
                sa.select([agus.c.user_id])
                .select_from(j)
                .where(groups.c.domain_name == domain_name)
            )
            result = await conn.execute(usr_query)
            usrs = result.fetchall()
            user_ids = [g.id for g in usrs]
            update_query = update_query.where(
                (error_logs.c.user.in_(user_ids)) & (error_logs.c.id == log_id),
            )
        else:
            update_query = update_query.where(
                (error_logs.c.user == user_uuid) & (error_logs.c.id == log_id),
            )

        result = await conn.execute(update_query)
        assert result.rowcount == 1

        return web.json_response({"success": True}, status=200)


async def log_cleanup_task(app: web.Application, src: AgentId, event: DoLogCleanupEvent) -> None:
    root_ctx: RootContext = app["_root.context"]
    etcd = root_ctx.shared_config.etcd
    raw_lifetime = await etcd.get("config/logs/error/retention")
    if raw_lifetime is None:
        raw_lifetime = "90d"
    try:
        lifetime = tx.TimeDuration().check(raw_lifetime)
    except ValueError:
        lifetime = dt.timedelta(days=90)
        log.warning(
            "Failed to parse the error log retention period ({}) read from etcd; "
            "falling back to 90 days",
            raw_lifetime,
        )
    boundary = datetime.now() - lifetime
    async with root_ctx.db.begin() as conn:
        query = sa.delete(error_logs).where(error_logs.c.created_at < boundary)
        result = await conn.execute(query)
        if result.rowcount > 0:
            log.info("Cleaned up {} log(s) filed before {}", result.rowcount, boundary)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    log_cleanup_timer: GlobalTimer
    log_cleanup_timer_redis: RedisConnectionInfo
    log_cleanup_timer_evh: EventHandler[web.Application, DoLogCleanupEvent]


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["logs.context"]
    app_ctx.log_cleanup_timer_evh = root_ctx.event_dispatcher.consume(
        DoLogCleanupEvent,
        app,
        log_cleanup_task,
    )
    app_ctx.log_cleanup_timer_redis = redis_helper.get_redis_object(
        root_ctx.shared_config.data["redis"],
        db=REDIS_LIVE_DB,
    )
    app_ctx.log_cleanup_timer = GlobalTimer(
        root_ctx.distributed_lock_factory(LockID.LOCKID_LOG_CLEANUP_TIMER, 20.0),
        root_ctx.event_producer,
        lambda: DoLogCleanupEvent(),
        20.0,
        initial_delay=17.0,
    )
    await app_ctx.log_cleanup_timer.join()


async def shutdown(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["logs.context"]
    await app_ctx.log_cleanup_timer.leave()
    root_ctx.event_dispatcher.unconsume(app_ctx.log_cleanup_timer_evh)
    await app_ctx.log_cleanup_timer_redis.close()


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (4, 5)
    app["prefix"] = "logs/error"
    app["logs.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "", append))
    cors.add(app.router.add_route("GET", "", list_logs))
    cors.add(app.router.add_route("POST", r"/{log_id}/clear", mark_cleared))

    return app, []



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Type

import yarl

if TYPE_CHECKING:
    from .base import BaseContainerRegistry


def get_container_registry(registry_info: Mapping[str, Any]) -> Type[BaseContainerRegistry]:
    registry_url = yarl.URL(registry_info[""])
    registry_type = registry_info.get("type", "docker")
    cr_cls: Type[BaseContainerRegistry]
    if registry_url.host is not None and registry_url.host.endswith(".docker.io"):
        from .docker import DockerHubRegistry

        cr_cls = DockerHubRegistry
    elif registry_type == "docker":
        from .docker import DockerRegistry_v2

        cr_cls = DockerRegistry_v2
    elif registry_type == "harbor":
        from .harbor import HarborRegistry_v1

        cr_cls = HarborRegistry_v1
    elif registry_type == "harbor2":
        from .harbor import HarborRegistry_v2

        cr_cls = HarborRegistry_v2
    elif registry_type == "local":
        from .local import LocalRegistry

        cr_cls = LocalRegistry
    else:
        raise RuntimeError(f"Unsupported registry type: {registry_type}")
    return cr_cls



from __future__ import annotations

import asyncio
import json
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Final,
    Iterable,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
)

import aiohttp_cors
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from aiohttp_sse import sse_response
from aiotools import adefer

from ai.backend.common import validators as tx
from ai.backend.common.events import (
    BgtaskCancelledEvent,
    BgtaskDoneEvent,
    BgtaskFailedEvent,
    BgtaskUpdatedEvent,
    EventDispatcher,
    KernelCancelledEvent,
    KernelCreatingEvent,
    KernelPreparingEvent,
    KernelPullingEvent,
    KernelStartedEvent,
    KernelTerminatedEvent,
    KernelTerminatingEvent,
    SessionCancelledEvent,
    SessionEnqueuedEvent,
    SessionFailureEvent,
    SessionScheduledEvent,
    SessionStartedEvent,
    SessionSuccessEvent,
    SessionTerminatedEvent,
    SessionTerminatingEvent,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import AgentId

from ..models import UserRole, groups, kernels
from ..models.utils import execute_with_retry
from ..types import Sentinel
from .auth import auth_required
from .exceptions import GenericForbidden, GroupNotFound, ObjectNotFound
from .manager import READ_ALLOWED, server_status_required
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext
    from .types import CORSOptions, WebMiddleware

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

sentinel: Final = Sentinel.token

SessionEventInfo = Tuple[str, dict, str, Optional[int]]
BgtaskEvents = Union[BgtaskUpdatedEvent, BgtaskDoneEvent, BgtaskCancelledEvent, BgtaskFailedEvent]


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["name", "sessionName"], default="*") >> "session_name": t.String,
            t.Key("ownerAccessKey", default=None) >> "owner_access_key": t.Null | t.String,
            t.Key("sessionId", default=None) >> "session_id": t.Null | tx.UUID,
            # NOTE: if set, sessionId overrides sessionName and ownerAccessKey parameters.
            tx.AliasedKey(["group", "groupName"], default="*") >> "group_name": t.String,
            t.Key("scope", default="*"): t.Enum("*", "session", "kernel"),
        }
    )
)
@adefer
async def push_session_events(
    defer,
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    app_ctx: PrivateContext = request.app["events.context"]
    session_name = params["session_name"]
    session_id = params["session_id"]
    scope = params["scope"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    access_key = params["owner_access_key"]
    if access_key is None:
        access_key = request["keypair"]["access_key"]
    if user_role == UserRole.USER:
        if access_key != request["keypair"]["access_key"]:
            raise GenericForbidden
    group_name = params["group_name"]
    my_queue: asyncio.Queue[Sentinel | SessionEventInfo] = asyncio.Queue()
    log.info("PUSH_SESSION_EVENTS (ak:{}, s:{}, g:{})", access_key, session_name, group_name)
    if group_name == "*":
        group_id = "*"
    else:
        async with root_ctx.db.begin_readonly() as conn:
            query = sa.select([groups.c.id]).select_from(groups).where(groups.c.name == group_name)
            result = await conn.execute(query)
            row = result.first()
            if row is None:
                raise GroupNotFound
            group_id = row["id"]
    app_ctx.session_event_queues.add(my_queue)
    defer(lambda: app_ctx.session_event_queues.remove(my_queue))
    async with sse_response(request) as resp:
        try:
            while True:
                evdata = await my_queue.get()
                try:
                    if evdata is sentinel:
                        break
                    event_name, row, reason, exit_code = evdata
                    if user_role in (UserRole.USER, UserRole.ADMIN):
                        if row["domain_name"] != request["user"]["domain_name"]:
                            continue
                    if user_role == UserRole.USER:
                        if row["user_uuid"] != user_uuid:
                            continue
                    if group_id != "*" and row["group_id"] != group_id:
                        continue
                    if scope == "session" and not event_name.startswith("session_"):
                        continue
                    if scope == "kernel" and not event_name.startswith("kernel_"):
                        continue
                    if session_id is not None:
                        if row["session_id"] != session_id:
                            continue
                    else:
                        if session_name != "*" and not (
                            (row["session_name"] == session_name)
                            and (row["access_key"] == access_key)
                        ):
                            continue
                    response_data = {
                        "reason": reason,
                        "sessionName": row["session_name"],
                        "ownerAccessKey": row["access_key"],
                        "sessionId": str(row["session_id"]),
                        "exitCode": exit_code,
                    }
                    if kernel_id := row.get("id"):
                        response_data["kernelId"] = str(kernel_id)
                    if cluster_role := row.get("cluster_role"):
                        response_data["clusterRole"] = cluster_role
                    if cluster_idx := row.get("cluster_idx"):
                        response_data["clusterIdx"] = cluster_idx
                    await resp.send(json.dumps(response_data), event=event_name)
                finally:
                    my_queue.task_done()
        finally:
            return resp


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["task_id", "taskId"]): tx.UUID,
        }
    )
)
async def push_background_task_events(
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    task_id = params["task_id"]
    access_key = request["keypair"]["access_key"]
    log.info("PUSH_BACKGROUND_TASK_EVENTS (ak:{}, t:{})", access_key, task_id)
    try:
        return await root_ctx.background_task_manager.push_bgtask_events(request, task_id)
    except ValueError as e:
        raise ObjectNotFound(extra_data=str(e), object_name="background task")


async def enqueue_kernel_creation_status_update(
    app: web.Application,
    source: AgentId,
    event: KernelPreparingEvent | KernelPullingEvent | KernelCreatingEvent | KernelStartedEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select(
                    [
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.cluster_role,
                        kernels.c.cluster_idx,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ]
                )
                .select_from(kernels)
                .where(
                    (kernels.c.id == event.kernel_id),
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.name, row._mapping, event.reason, None))


async def enqueue_kernel_termination_status_update(
    app: web.Application,
    agent_id: AgentId,
    event: KernelCancelledEvent | KernelTerminatingEvent | KernelTerminatedEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select(
                    [
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.cluster_role,
                        kernels.c.cluster_idx,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ]
                )
                .select_from(kernels)
                .where(
                    (kernels.c.id == event.kernel_id),
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        exit_code = (
            event.exit_code
            if isinstance(event, (KernelTerminatingEvent, KernelTerminatedEvent))
            else None
        )
        q.put_nowait((event.name, row._mapping, event.reason, exit_code))


async def enqueue_session_creation_status_update(
    app: web.Application,
    source: AgentId,
    event: SessionEnqueuedEvent
    | SessionScheduledEvent
    | SessionStartedEvent
    | SessionCancelledEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select(
                    [
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ]
                )
                .select_from(kernels)
                .where(
                    (kernels.c.id == event.session_id),
                    # for the main kernel, kernel ID == session ID
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.name, row._mapping, event.reason, None))


async def enqueue_session_termination_status_update(
    app: web.Application,
    agent_id: AgentId,
    event: SessionTerminatingEvent | SessionTerminatedEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select(
                    [
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ]
                )
                .select_from(kernels)
                .where(
                    (kernels.c.session_id == event.session_id),
                    # for the main kernel, kernel ID == session ID
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.name, row._mapping, event.reason, None))


async def enqueue_batch_task_result_update(
    app: web.Application,
    agent_id: AgentId,
    event: SessionSuccessEvent | SessionFailureEvent,
) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]

    async def _fetch():
        async with root_ctx.db.begin_readonly() as conn:
            query = (
                sa.select(
                    [
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ]
                )
                .select_from(kernels)
                .where(
                    (kernels.c.session_id == event.session_id),
                )
            )
            result = await conn.execute(query)
            return result.first()

    row = await execute_with_retry(_fetch)
    if row is None:
        return
    for q in app_ctx.session_event_queues:
        q.put_nowait((event.name, row._mapping, event.reason, event.exit_code))


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    session_event_queues: Set[asyncio.Queue[Sentinel | SessionEventInfo]]


async def events_app_ctx(app: web.Application) -> AsyncIterator[None]:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["events.context"]
    app_ctx.session_event_queues = set()
    event_dispatcher: EventDispatcher = root_ctx.event_dispatcher
    event_dispatcher.subscribe(SessionEnqueuedEvent, app, enqueue_session_creation_status_update)
    event_dispatcher.subscribe(SessionScheduledEvent, app, enqueue_session_creation_status_update)
    event_dispatcher.subscribe(KernelPreparingEvent, app, enqueue_kernel_creation_status_update)
    event_dispatcher.subscribe(KernelPullingEvent, app, enqueue_kernel_creation_status_update)
    event_dispatcher.subscribe(KernelCreatingEvent, app, enqueue_kernel_creation_status_update)
    event_dispatcher.subscribe(KernelStartedEvent, app, enqueue_kernel_creation_status_update)
    event_dispatcher.subscribe(SessionStartedEvent, app, enqueue_session_creation_status_update)
    event_dispatcher.subscribe(
        KernelTerminatingEvent, app, enqueue_kernel_termination_status_update
    )
    event_dispatcher.subscribe(KernelTerminatedEvent, app, enqueue_kernel_termination_status_update)
    event_dispatcher.subscribe(KernelCancelledEvent, app, enqueue_kernel_termination_status_update)
    event_dispatcher.subscribe(
        SessionTerminatingEvent, app, enqueue_session_termination_status_update
    )
    event_dispatcher.subscribe(
        SessionTerminatedEvent, app, enqueue_session_termination_status_update
    )
    event_dispatcher.subscribe(SessionCancelledEvent, app, enqueue_session_creation_status_update)
    event_dispatcher.subscribe(SessionSuccessEvent, app, enqueue_batch_task_result_update)
    event_dispatcher.subscribe(SessionFailureEvent, app, enqueue_batch_task_result_update)
    root_ctx.background_task_manager.register_event_handlers(event_dispatcher)
    yield


async def events_shutdown(app: web.Application) -> None:
    # shutdown handler is called before waiting for closing active connections.
    # We need to put sentinels here to ensure delivery of them to active SSE connections.
    app_ctx: PrivateContext = app["events.context"]
    join_tasks = []
    for sq in app_ctx.session_event_queues:
        sq.put_nowait(sentinel)
        join_tasks.append(sq.join())
    await asyncio.gather(*join_tasks)


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "events"
    app["events.context"] = PrivateContext()
    app["api_versions"] = (3, 4)
    app.on_shutdown.append(events_shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    app.cleanup_ctx.append(events_app_ctx)
    cors.add(add_route("GET", r"/background-task", push_background_task_events))
    cors.add(add_route("GET", r"/session", push_session_events))
    return app, []



"""rename-clone_allowed-to-cloneable

Revision ID: 57e717103287
Revises: eec98e65902a
Create Date: 2020-10-04 14:14:55.167654

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "57e717103287"
down_revision = "eec98e65902a"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("vfolders", "clone_allowed", new_column_name="cloneable")


def downgrade():
    op.alter_column("vfolders", "cloneable", new_column_name="clone_allowed")



"""remove_keypair_concurrency_used

Revision ID: d59ff89e7514
Revises: 0f7a4b643940
Create Date: 2022-03-21 16:43:29.899251

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d59ff89e7514"
down_revision = "0f7a4b643940"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_column("keypairs", "concurrency_used")


def downgrade():
    op.add_column(
        "keypairs",
        sa.Column(
            "concurrency_used",
            sa.Integer,
            nullable=True,
            default=0,
            server_default="0",
        ),
    )



"""add is_local to images

Revision ID: 10c58e701d87
Revises: 36b69ddee76e
Create Date: 2023-02-24 06:28:41.583932

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "10c58e701d87"
down_revision = "36b69ddee76e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "images",
        sa.Column("is_local", sa.Boolean(), server_default=sa.text("false"), nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("images", "is_local")
    # ### end Alembic commands ###



from __future__ import annotations

from decimal import Decimal
from typing import List, Optional, Sequence, Tuple

import trafaret as t

from ai.backend.common.types import AgentId, ResourceSlot, SessionId

from ..models import AgentRow, SessionRow
from .types import AbstractScheduler, KernelInfo


def key_by_requested_slots(
    agent: AgentRow,
    requested_slots: ResourceSlot,
) -> Tuple[int, ResourceSlot]:
    unused_slot_keys = set()
    for k, v in requested_slots.items():
        if v == Decimal(0):
            unused_slot_keys.add(k)
    num_extras = 0
    for k, v in agent.available_slots.items():
        if k in unused_slot_keys and v > Decimal(0):
            num_extras += 1
    # Put back agents with more extra slot types
    # (e.g., accelerators)
    # Also put front agents with exactly required slot types
    return (-num_extras, agent.available_slots)


class FIFOSlotScheduler(AbstractScheduler):
    config_iv = t.Dict(
        {
            t.Key("num_retries_to_skip", default=0): t.ToInt(gte=0),
        }
    ).allow_extra("*")

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[SessionRow],
        existing_sessions: Sequence[SessionRow],
    ) -> Optional[SessionId]:
        local_pending_sessions = list(pending_sessions)
        skipped_sessions: List[SessionRow] = []
        max_retries = self.config["num_retries_to_skip"]
        while local_pending_sessions:
            # Just pick the first pending session, but skip it
            # if it has more than 3 failures.
            s = local_pending_sessions.pop(0)
            if max_retries == 0:  # it's strict FIFO
                return s.id
            if s.status_data is not None:
                sched_data = s.status_data.get("scheduler", {})
                if sched_data.get("retries", 0) >= max_retries:
                    skipped_sessions.append(s)
                    continue
            return s.id
        # But if all sessions are skipped, then choose the first one.
        if skipped_sessions:
            return skipped_sessions[0].id
        return None

    def _assign_agent(
        self,
        agents: Sequence[AgentRow],
        requested_slots: ResourceSlot,
    ) -> Optional[AgentId]:
        possible_agents = []
        for agent in agents:
            remaining_slots = agent.available_slots - agent.occupied_slots
            if remaining_slots >= requested_slots:
                possible_agents.append(agent)
        if possible_agents:
            chosen_agent = max(
                possible_agents,
                key=lambda a: key_by_requested_slots(
                    a,
                    requested_slots,
                ),
            )
            return chosen_agent.id
        return None

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentRow],
        pending_session: SessionRow,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_session.requested_slots,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentRow],
        pending_kernel: KernelInfo,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_kernel.requested_slots,
        )


class LIFOSlotScheduler(AbstractScheduler):
    config_iv = t.Dict({}).allow_extra("*")

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[SessionRow],
        existing_sessions: Sequence[SessionRow],
    ) -> Optional[SessionId]:
        # Just pick the last pending session.
        return SessionId(pending_sessions[-1].id)

    def _assign_agent(
        self,
        agents: Sequence[AgentRow],
        requested_slots: ResourceSlot,
    ) -> Optional[AgentId]:
        possible_agents = []
        for agent in agents:
            remaining_slots = agent.available_slots - agent.occupied_slots
            if remaining_slots >= requested_slots:
                possible_agents.append(agent)
        if possible_agents:
            chosen_agent = max(
                possible_agents,
                key=lambda a: key_by_requested_slots(
                    a,
                    requested_slots,
                ),
            )
            return chosen_agent.id
        return None

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentRow],
        pending_session: SessionRow,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_session.requested_slots,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentRow],
        pending_kernel: KernelInfo,
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_kernel.requested_slots,
        )



"""add_session_table

Revision ID: b6b884fbae1f
Revises: 213a04e90ecf
Create Date: 2022-12-05 16:12:53.275671

"""
from collections import defaultdict
from uuid import uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.orm import registry

from ai.backend.manager.models import KernelStatus, SessionStatus
from ai.backend.manager.models.base import GUID, KernelIDColumn, convention

# revision identifiers, used by Alembic.
revision = "b6b884fbae1f"
down_revision = "213a04e90ecf"
branch_labels = None
depends_on = None


kernel_status = tuple(s.name for s in KernelStatus)

session_status = tuple(s.name for s in SessionStatus)

metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base = mapper_registry.generate_base()


def default_hostname(context) -> str:
    params = context.get_current_parameters()
    return f"{params['cluster_role']}{params['cluster_idx']}"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    # db_session = Session(connection)

    # class ImageRow(Base):
    #     __tablename__ = "images"
    #     id = sa.Column("id", GUID(), nullable=False, primary_key=True)
    #     name = sa.Column("name", sa.String, nullable=False, index=True)

    kernels = sa.Table(
        "kernels",
        metadata,
        KernelIDColumn(),
        sa.Column("session_id", GUID, nullable=False),
        sa.Column("session_creation_id", sa.String(length=32), unique=False),
        sa.Column("session_name", sa.String(length=64), unique=False),
        sa.Column(
            "session_type",
            pgsql.ENUM("INTERACTIVE", "BATCH"),
            nullable=False,
            default="INTERACTIVE",
            server_default="INTERACTIVE",
        ),
        sa.Column("cluster_role", sa.String(length=16), nullable=False, default="main"),
        sa.Column(
            "cluster_mode",
            sa.String(length=16),
            nullable=False,
            default="SINGLE_NODE",
            server_default="SINGLE_NODE",
        ),
        sa.Column("cluster_size", sa.Integer, nullable=False, default=1),
        # Resource ownership
        sa.Column("scaling_group", sa.ForeignKey("scaling_groups.name"), nullable=True),
        sa.Column(
            "domain_name", sa.String(length=64), sa.ForeignKey("domains.name"), nullable=False
        ),
        sa.Column("group_id", GUID, sa.ForeignKey("groups.id"), nullable=False),
        sa.Column("user_uuid", GUID, sa.ForeignKey("users.uuid"), nullable=False),
        sa.Column("access_key", sa.String(length=20), sa.ForeignKey("keypairs.access_key")),
        sa.Column("image", sa.String(length=512)),
        sa.Column("tag", sa.String(length=64), nullable=True),
        # Resource occupation
        sa.Column("occupied_slots", pgsql.JSONB(), nullable=False),
        sa.Column("vfolder_mounts", pgsql.JSONB(), nullable=True),
        sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={}),
        sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True),
        sa.Column("use_host_network", sa.Boolean(), default=False, nullable=False),
        # Lifecycle
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
        sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True, default=sa.null()),
        sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True, default=sa.null()),
        sa.Column(
            "status",
            sa.Enum(*kernel_status),
            default="PENDING",
            server_default="PENDING",
            nullable=False,
        ),
        # sa.Column("status_changed", sa.DateTime(timezone=True), nullable=True),
        sa.Column("status_history", pgsql.JSONB(), nullable=True, default=sa.null()),
        sa.Column("status_info", sa.Unicode(), nullable=True, default=sa.null()),
        sa.Column("status_data", pgsql.JSONB(), nullable=True, default=sa.null()),
        sa.Column("callback_url", sa.types.UnicodeText, nullable=True, default=sa.null()),
        sa.Column("startup_command", sa.Text, nullable=True),
        sa.Column(
            "result",
            sa.Enum("UNDEFINED", "SUCCESS", "FAILURE"),
            default="UNDEFINED",
            server_default="UNDEFINED",
            nullable=False,
        ),
        # Resource metrics measured upon termination
        sa.Column("num_queries", sa.BigInteger(), default=0),
        sa.Column("last_stat", pgsql.JSONB(), nullable=True, default=sa.null()),
    )

    class SessionRow(Base):
        __tablename__ = "sessions"
        id = sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
            primary_key=True,
        )
        creation_id = sa.Column("creation_id", sa.String(length=32), unique=False)
        name = sa.Column("name", sa.String(length=64), unique=False)
        session_type = sa.Column(
            "session_type",
            pgsql.ENUM("INTERACTIVE", "BATCH", name="sessiontypes", create_type=False),
            server_default="INTERACTIVE",
            nullable=False,
        )
        cluster_mode = sa.Column(
            "cluster_mode", sa.String(length=16), server_default="SINGLE_NODE", nullable=False
        )
        cluster_size = sa.Column("cluster_size", sa.Integer(), nullable=False)

        # Resource ownership
        scaling_group_name = sa.Column("scaling_group_name", sa.String(length=64), nullable=True)
        target_sgroup_names = sa.Column(
            "target_sgroup_names",
            sa.ARRAY(sa.String(length=64)),
            server_default="{}",
            nullable=True,
        )
        domain_name = sa.Column("domain_name", sa.String(length=64), nullable=False)
        group_id = sa.Column("group_id", GUID(), nullable=False)
        user_uuid = sa.Column("user_uuid", GUID(), nullable=False)
        access_key = sa.Column("access_key", sa.String(length=20), nullable=True)
        # kp_resource_policy = sa.Column(
        #     'kp_resource_policy', sa.String(length=256), nullable=False)

        # # if image_id is null, should find a image field from related kernel row.
        # image_id = sa.Column("image_id", GUID(), nullable=True)
        # # `image` column is identical to kernels `image` column.
        # image = sa.Column("image", sa.String(length=512))
        tag = sa.Column("tag", sa.String(length=64), nullable=True)

        # Resource occupation
        occupying_slots = sa.Column("occupying_slots", pgsql.JSONB(), nullable=False)
        requested_slots = sa.Column("requested_slots", pgsql.JSONB(), nullable=False)
        vfolder_mounts = sa.Column("vfolder_mounts", pgsql.JSONB(), nullable=True)
        resource_opts = sa.Column("resource_opts", pgsql.JSONB(), nullable=True, default={})
        environ = sa.Column("environ", pgsql.JSONB(), nullable=True)
        bootstrap_script = sa.Column("bootstrap_script", sa.String(length=16 * 1024), nullable=True)
        use_host_network = sa.Column(
            "use_host_network", sa.Boolean(), default=False, nullable=False
        )

        # Lifecycle
        timeout = sa.Column("timeout", sa.BigInteger(), nullable=True)
        created_at = sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        )
        terminated_at = sa.Column("terminated_at", sa.DateTime(timezone=True), nullable=True)
        starts_at = sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True)
        status = sa.Column(
            "status",
            pgsql.ENUM(
                *session_status,
                name="sessionstatus",
                create_type=True,
            ),
            server_default="PENDING",
            nullable=False,
        )
        # status_changed = sa.Column("status_changed", sa.DateTime(timezone=True), nullable=True)
        status_history = sa.Column(
            "status_history", pgsql.JSONB(), nullable=True, default=sa.null()
        )
        status_info = sa.Column("status_info", sa.Unicode(), nullable=True)
        status_data = sa.Column("status_data", pgsql.JSONB(astext_type=sa.Text()), nullable=True)
        callback_url = sa.Column("callback_url", sa.types.UnicodeText(), nullable=True)

        startup_command = sa.Column("startup_command", sa.Text(), nullable=True)
        result = sa.Column(
            "result",
            pgsql.ENUM("UNDEFINED", "SUCCESS", "FAILURE", name="sessionresults", create_type=True),
            server_default="UNDEFINED",
            nullable=False,
        )

        # Resource metrics measured upon termination
        num_queries = sa.Column("num_queries", sa.BigInteger(), nullable=True)
        last_stat = sa.Column("last_stat", pgsql.JSONB(astext_type=sa.Text()), nullable=True)

    SessionRow.__table__.create(connection)
    op.create_foreign_key(
        op.f("fk_sessions_access_key_keypairs"),
        "sessions",
        "keypairs",
        ["access_key"],
        ["access_key"],
    )
    op.create_foreign_key(
        op.f("fk_sessions_domain_name_domains"), "sessions", "domains", ["domain_name"], ["name"]
    )
    op.create_foreign_key(
        op.f("fk_sessions_group_id_groups"), "sessions", "groups", ["group_id"], ["id"]
    )
    # op.create_foreign_key(
    #     op.f("fk_sessions_image_images"), "sessions", "images", ["image_id"], ["id"]
    # )
    op.create_foreign_key(
        op.f("fk_sessions_scaling_group_scaling_groups"),
        "sessions",
        "scaling_groups",
        ["scaling_group_name"],
        ["name"],
    )
    op.create_foreign_key(
        op.f("fk_sessions_user_uuid_users"), "sessions", "users", ["user_uuid"], ["uuid"]
    )
    op.create_index(op.f("ix_sessions_created_at"), "sessions", ["created_at"], unique=False)
    op.create_index(op.f("ix_sessions_name"), "sessions", ["name"], unique=False)
    op.create_index(op.f("ix_sessions_result"), "sessions", ["result"], unique=False)
    op.create_index(
        op.f("ix_sessions_scaling_group_name"), "sessions", ["scaling_group_name"], unique=False
    )
    op.create_index(op.f("ix_sessions_session_type"), "sessions", ["session_type"], unique=False)
    # op.create_index(op.f("ix_sessions_status"), "sessions", ["status"], unique=False)
    # op.create_index(
    #     op.f("ix_sessions_status_changed"), "sessions", ["status_changed"], unique=False
    # )
    op.create_index(op.f("ix_sessions_terminated_at"), "sessions", ["terminated_at"], unique=False)

    query = sa.select([kernels]).select_from(kernels).order_by(kernels.c.created_at)
    kernel_rows = connection.execute(query).fetchall()

    # imgs = db_session.query(ImageRow).all()
    # img_map = {img.name: img.id for img in imgs}

    all_kernel_sessions = {}
    # kernel_id and session_id are the same if it is a single kernel session.
    # update kernel's session_id of single kernel session.
    single_kernel_ids = {}
    for row in kernel_rows:
        sess_id = row["session_id"]
        if sess_id not in all_kernel_sessions:
            if sess_id == row["id"]:
                # create session with new id if the kernel_id and session_id are the same.
                sess_id = uuid4()
                single_kernel_ids[row["id"]] = sess_id
            sess = {
                "id": sess_id,
                "creation_id": row["session_creation_id"],
                "name": row["session_name"],
                "session_type": row["session_type"],
                "cluster_mode": row["cluster_mode"],
                "cluster_size": row["cluster_size"],
                "scaling_group_name": row["scaling_group"],
                "domain_name": row["domain_name"],
                "group_id": row["group_id"],
                "user_uuid": row["user_uuid"],
                "access_key": row["access_key"],
                # "image_id": img_map[row["image"]],
                # "image": row["image"],
                "tag": row["tag"],
                "occupying_slots": row["occupied_slots"],
                "requested_slots": row["occupied_slots"],
                "vfolder_mounts": row["vfolder_mounts"],
                "resource_opts": row["resource_opts"],
                "bootstrap_script": row["bootstrap_script"],
                "use_host_network": row["use_host_network"],
                "created_at": row["created_at"],
                "terminated_at": row["terminated_at"],
                "starts_at": row["starts_at"],
                "status": row["status"],
                # "status_changed": row["status_changed"],
                "status_history": row["status_history"],
                "status_info": row["status_info"],
                "status_data": row["status_data"],
                "callback_url": row["callback_url"],
                "startup_command": row["startup_command"],
                "result": row["result"],
                "num_queries": row["num_queries"],
                "last_stat": row["last_stat"],
            }
            all_kernel_sessions[sess_id] = sess
        else:
            sess = all_kernel_sessions[sess_id]

            # st_change = sess["status_changed"]

            try:
                if sess["created_at"] is not None and row["created_at"] is not None:
                    if sess["created_at"] > row["created_at"]:
                        sess["created_at"] = row["created_at"]
                        # st_change = row["st_change"]
            except (TypeError, ValueError):
                pass

            try:
                if sess["starts_at"] is not None and row["starts_at"] is not None:
                    if sess["starts_at"] > row["starts_at"]:
                        sess["starts_at"] = row["starts_at"]
                        # st_change = row["st_change"]
            except (TypeError, ValueError):
                pass

            # if str(sess["status"]) not in ("ERROR", "CANCELLED"):
            #     if KernelStatus.index(str(sess["status"])) > KernelStatus.index(str(row["status"])):
            #         sess["status"] = row["status"]
            sess["status"] = row["status"]

            try:
                if sess["terminated_at"] is not None and row["terminated_at"] is not None:
                    if sess["terminated_at"] < row["terminated_at"]:
                        sess["terminated_at"] = row["terminated_at"]
                        # st_change = row["status_changed"]
                elif sess["terminated_at"] is None:
                    pass
                elif row["terminated_at"] is None:
                    sess["terminated_at"] = None
                # sess["status_changed"] = st_change
            except (TypeError, ValueError):
                pass

            if sess["status_info"] != row["status_info"]:
                import json

                sinfo = sess["status_info"]
                try:
                    sinfo = json.loads(sinfo)
                except json.decoder.JSONDecodeError:
                    sinfo = {sess["id"]: sinfo}
                sinfo = {**sinfo, row["id"]: row["status_info"]}
                sess["status_info"] = json.dumps(sinfo)

    if all_kernel_sessions:
        creates = tuple(all_kernel_sessions.values())
        connection.execute(SessionRow.__table__.insert(), creates)

    # Session dependency table
    op.create_foreign_key(
        op.f("fk_session_dependencies_session_id_sessions"),
        "session_dependencies",
        "sessions",
        ["session_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        op.f("fk_session_dependencies_depends_on_sessions"),
        "session_dependencies",
        "sessions",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.drop_constraint(
        "fk_session_dependencies_depends_on_kernels", "session_dependencies", type_="foreignkey"
    )
    op.drop_constraint(
        "fk_session_dependencies_session_id_kernels", "session_dependencies", type_="foreignkey"
    )

    # Kernel table
    sess_query = (
        sa.update(kernels)
        .where(kernels.c.id == sa.bindparam("kernel_id"))
        .values(session_id=sa.bindparam("b_session_id"))
    )
    sess_update_params = []
    for kern_id, sess_id in single_kernel_ids.items():
        sess_update_params.append(
            {
                "kernel_id": kern_id,
                "b_session_id": sess_id,
            }
        )
    if sess_update_params:
        connection.execute(sess_query, sess_update_params)
    op.create_foreign_key(
        op.f("fk_kernels_session_id_sessions"), "kernels", "sessions", ["session_id"], ["id"]
    )
    # op.add_column("kernels", sa.Column("image_id", GUID(), nullable=True))
    # kernels = sa.Table(
    #     "kernels",
    #     metadata,
    #     KernelIDColumn(),
    #     sa.Column("image_id", GUID(), nullable=True),
    #     extend_existing=True,
    # )
    # img_query = (
    #     sa.update(kernels)
    #     .where(kernels.c.id == sa.bindparam("kernel_id"))
    #     .values(image_id=sa.bindparam("b_image_id"))
    # )
    # img_update_params = []
    # for kern in kernel_rows:
    #     try:
    #         img_update_params.append(
    #             {
    #                 "kernel_id": kern["id"],
    #                 "b_image_id": img_map[kern["image"]],
    #             }
    #         )
    #     except KeyError:
    #         continue
    # if img_update_params:
    #     connection.execute(img_query, img_update_params)

    # op.create_foreign_key(
    #     op.f("fk_kernels_image_id_images"), "kernels", "images", ["image_id"], ["id"]
    # )
    op.add_column("kernels", sa.Column("requested_slots", pgsql.JSONB(), nullable=True))

    ## Association tables
    # groups_users
    op.drop_constraint(
        "uq_association_user_id_group_id", "association_groups_users", type_="unique"
    )

    # scaling_groups_domains
    op.drop_constraint("uq_sgroup_domain", "sgroups_for_domains", type_="unique")

    # scaling_groups_groups
    op.drop_constraint("uq_sgroup_ugroup", "sgroups_for_groups", type_="unique")

    # scaling_groups_keypairs
    op.drop_constraint("uq_sgroup_akey", "sgroups_for_keypairs", type_="unique")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # class ImageRow(Base):
    #     __tablename__ = "images"
    #     id = sa.Column("id", GUID(), nullable=False, primary_key=True)
    #     name = sa.Column("name", sa.String, nullable=False, index=True)

    connection = op.get_bind()
    # db_session = Session(connection)
    # imgs = db_session.query(ImageRow).all()
    # img_map = {img.id: img.name for img in imgs}

    # Kernel table
    op.drop_constraint(op.f("fk_kernels_session_id_sessions"), "kernels", type_="foreignkey")
    # op.drop_constraint(op.f("fk_kernels_image_id_images"), "kernels", type_="foreignkey")
    kernels = sa.Table(
        "kernels",
        metadata,
        KernelIDColumn(),
        sa.Column("session_id", GUID, nullable=False),
        # sa.Column("image_id", GUID(), nullable=True),
        # sa.Column("image", sa.String(length=512), nullable=True),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    )
    query = sa.select(kernels).select_from(kernels).order_by(kernels.c.created_at)
    kernel_rows = connection.execute(query).fetchall()

    # img_update_query = (
    #     sa.update(kernels)
    #     .where(kernels.c.id == sa.bindparam("kernel_id"))
    #     .values({kernels.c.image: sa.bindparam("image")})
    # )
    # img_params = []
    num_kernel_sess = defaultdict(int)
    sess_kern_map = {}
    for kern in kernel_rows:
        # restore kernel's session_id
        sess_id = kern["session_id"]
        num_kernel_sess[sess_id] += 1
        if sess_id not in sess_kern_map:
            sess_kern_map[sess_id] = kern["id"]
    #     try:
    #         img_params.append(
    #             {
    #                 "kernel_id": kern["id"],
    #                 "image": img_map[kern["image_id"]],
    #             }
    #         )
    #     except KeyError:
    #         continue
    # if img_params:
    #     connection.execute(img_update_query, img_params)
    kern_update_query = (
        sa.update(kernels)
        .where(kernels.c.id == sa.bindparam("kernel_id"))
        .values({kernels.c.session_id: sa.bindparam("session_id")})
    )
    sess_update_params = []
    for sess_id, num_kern in num_kernel_sess.items():
        if num_kern == 1:
            # it is single kernel session.
            # session_id must be same with kernel_id.
            kernel_id = sess_kern_map[sess_id]
            sess_update_params.append(
                {
                    "kernel_id": kernel_id,
                    "session_id": kernel_id,
                }
            )
    if sess_update_params:
        connection.execute(kern_update_query, sess_update_params)
    # op.drop_column("kernels", "image_id")
    op.drop_column("kernels", "requested_slots")

    # Session dependency table
    op.create_foreign_key(
        "fk_session_dependencies_session_id_kernels",
        "session_dependencies",
        "kernels",
        ["session_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "fk_session_dependencies_depends_on_kernels",
        "session_dependencies",
        "kernels",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.drop_constraint(
        op.f("fk_session_dependencies_depends_on_sessions"),
        "session_dependencies",
        type_="foreignkey",
    )
    op.drop_constraint(
        op.f("fk_session_dependencies_session_id_sessions"),
        "session_dependencies",
        type_="foreignkey",
    )

    # Session table
    op.drop_index(op.f("ix_sessions_terminated_at"), table_name="sessions")
    # op.drop_index(op.f("ix_sessions_status_changed"), table_name="sessions")
    # op.drop_index(op.f("ix_sessions_status"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_session_type"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_scaling_group_name"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_result"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_name"), table_name="sessions")
    op.drop_index(op.f("ix_sessions_created_at"), table_name="sessions")
    op.drop_table("sessions")

    # Drop ENUM types
    sess_statues = pgsql.ENUM(
        *session_status,
        name="sessionstatus",
    )
    sess_results = pgsql.ENUM("UNDEFINED", "SUCCESS", "FAILURE", name="sessionresults")

    sess_statues.drop(op.get_bind())
    sess_results.drop(op.get_bind())

    ## Association tables
    # groups users
    op.create_unique_constraint(
        "uq_association_user_id_group_id", "association_groups_users", ["user_id", "group_id"]
    )

    # scaling groups domains
    op.create_unique_constraint(
        "uq_sgroup_domain", "sgroups_for_domains", ["scaling_group", "domain"]
    )

    # scaling_groups_groups
    op.create_unique_constraint(
        "uq_sgroup_ugroup", "sgroups_for_groups", ["scaling_group", "group"]
    )

    # scaling_groups_keypairs
    op.create_unique_constraint(
        "uq_sgroup_akey", "sgroups_for_keypairs", ["scaling_group", "access_key"]
    )

    ### end Alembic commands ###



"""add_bootstrap_script_to_keypairs

Revision ID: ed666f476f39
Revises: d643752544de
Create Date: 2020-03-15 17:40:46.754121

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ed666f476f39"
down_revision = "d643752544de"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "keypairs",
        sa.Column(
            "bootstrap_script", sa.String(length=64 * 1024), nullable=False, server_default=""
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("keypairs", "bootstrap_script")
    # ### end Alembic commands ###
